#include <boost/random.hpp>
#include <boost/random/normal_distribution.hpp>
#include "dataset.h"
#include "common.h"
#include <assert.h>
#include <iostream>
#include <sstream>
#include <stdlib.h>
#include <sys/stat.h>
#include <errno.h>

using namespace boost;
using namespace std;

DataSet::DataSet()
{
}


void DataSet::randomWalks(long N, long length) {
    mt19937 rng;
    uniform_real<> ur(0.0,1.0);
    variate_generator<mt19937&,uniform_real<> > drawSample(rng, ur);

    this->N = N;

    for(int i = 0; i < N ; i++) {        
        originalTimeSeries[i].resize(length);
        for(int j = 0; j < length; j++) {
            originalTimeSeries[i][j] = 100.0;            
            for(int k = 0; k < j; k++)
                originalTimeSeries[i][j]+= drawSample()-0.5;
        }
    }

}

void DataSet::readFromFile(const char *src) {
    ifstream fin(src);

    cerr << "Reading real timeseries collection from file " << src << "...\n";

    string line;

    int i = 0;

    while (std::getline(fin, line)) {

        if(i >= TIMESERIES_MAX) {
            FATAL("too many timeseries in this collection, you should increase TIMESERIES_MAX");
        }

        stringstream lineStream(line);

        double v;

        while(! lineStream.eof()) {
            lineStream >> v;
            if(!lineStream.eof()) {
                originalTimeSeries[i].push_back(v);
            }
        }

        normalize(originalTimeSeries[i]);
        cerr << "TimeSeries " << i+1 <<  " has length " << originalTimeSeries[i].size() << "\n";
        i++;
    }

    N = i;

    cerr << "Done.\n";

    fin.close();

}


void DataSet::normalize(vector<double> &timeSeries) {
    double mean = 0;
    double maxdev = 0;

    //cerr << "Normalizing timeseries...\n";
    for(unsigned int i = 0; i < timeSeries.size(); i++) {
        mean += timeSeries[i];
    }
    mean /= (double)timeSeries.size();

    for(unsigned int i = 0; i < timeSeries.size(); i++) {

        maxdev = MAX(maxdev, ABS(mean - timeSeries[i]));
    }
    //cerr << "mean=" << mean << " maxdev=" << maxdev <<  "\n";
    for(unsigned int i = 0; i < timeSeries.size(); i++) {
        timeSeries[i] = (timeSeries[i] - mean) / maxdev;
    }
}


void DataSet::writeMultiSamplesDir(const char *dst) {

    cerr << "Writing all timeseries to separate files in directory " << dst << " ...\n";

    for(int i = 0; i < N; i++) {
        for(int j = 0; j < SAMPLES_MAX; j++) {
            normalize(perturbatedTimeSeries[i][j]);
        }
    }


    if(mkdir(dst,S_IRUSR|S_IWUSR|S_IXUSR) != 0) {
        FATAL("mkdir: " << strerror(errno));
    }

     char name[512];

    for( int i = 0; i < N; i++) {

        sprintf(name, "%s/timeseries%d", dst,i+1);

        ofstream fout(name);

        if(!fout) {
            FATAL("error opening outputfile " << name << ": " << strerror(errno));
        }

        fout.precision(4);

        for(unsigned int k = 0; k < originalTimeSeries[i].size(); k++) {
            for(int j = 0; j < SAMPLES_MAX; j++) {
                fout << perturbatedTimeSeries[i][j][k] << (j == SAMPLES_MAX-1 ? "" : " ");
            }
            fout << endl;
        }

        fout.close();
    }

    cerr << "Done.\n" ;

}


void DataSet::perturbateNormal(double mu, double sigma) {
    mt19937 rng(static_cast<unsigned> (time(0)));
    normal_distribution<> nd(mu, sigma);

    variate_generator<mt19937&,normal_distribution<> > drawSample(rng, nd);

    cerr << "adding perturbation (normal) ...\n";

    for( int i = 0; i < N; i++) {

        cerr << "processing TimeSeries " << i+1 << " ...\n";

        for(unsigned int j = 0; j < SAMPLES_MAX; j++) {
            perturbatedTimeSeries[i][j].resize(originalTimeSeries[i].size());
        }

        for( int i = 0; i < N; i++) {
            for(int j = 0; j < SAMPLES_MAX; j++) {
                for(unsigned int k = 0; k < perturbatedTimeSeries[i][j].size(); k++) {
                    perturbatedTimeSeries[i][j][k] = originalTimeSeries[i][k] + drawSample();
                }
            }
        }
    }
}
