/* includes */

#include <limits>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <iostream>

#include "main.h"
#include "dataset.h"
#include "common.h"
#include "timeseries.h"
#include "timeseriescollection.h"
#include "euclidean.h"
#include "precisionrecallM.h"
#include "movingeuclidean.h"
#include <boost/random.hpp>
using namespace boost;


using namespace std;

/* globals */
OPT o;

#define LOOKUPTABLES_PATHNAME "lookuptablesMixed"

const char *distribs[] = {"uniform", "normal", "exponential", NULL};

const char *names[] = {"50words",
                       "Adiac",
                       "Beef",
                       "CBF",
                       "Coffee",
                       "ECG200",
                       "FaceAll",
                       "FaceFour",
                       "FISH",
                       "Gun_Point",
                       "Lighting2",
                       "Lighting7",
                       "OliveOil",
                       "OSULeaf",
                       "SwedishLeaf",
                       "synthetic_control",
                       "Trace",
                       NULL};


bool active[] = {
    false, //    "50words",
    false, //   "Adiac",
    false, //   "Beef",
    false, //   "CBF",
    false, //   "Coffee",
    false, //   "ECG200",
    false, //   "FaceAll",
    false, //   "FaceFour",
    false, //   "FISH",
    true, //   "Gun_Point",
    false, //    "Lighting2",
    false, //    "Lighting7",
    false, //   "OliveOil",
    false, //  "OSULeaf",
    false, //   "SwedishLeaf",
    false,  //   "synthetic_control",
    false //  "Trace", // 17 datasets
};



/* prototypes */
int main(int argc, char **argv);
void initOpt(int argc, char **argv);
void help();
void cleanUp();
void experiment6(double stddev0, double stddev1);
void experiment7();
void printResultSet(char *name, vector<int> &v);
void printPrecisionRecall(char *name, PrecisionRecall &pr);
void generateFDustTables();

/* extern */


/*******************************************/


int main(int argc, char **argv) {
    initOpt(argc, argv);
    
    cout.precision(4);
    cerr.precision(4);
    
    
      // experiment6(0.4, 1.0);

    experiment7();

    //  experiment6(1.0, 2.0);
    
    cleanUp();
}



void experiment6(double stddev0, double stddev1) {
    char inputPathname[1024];
    char outputPathname[1024];
    
    for(int d = 0; distribs[d]; d++) {
        
        if(d != 1) // if not normal distrib
            continue;
        
        sprintf(outputPathname, "results/%s_exactquery2_mixed_%.1f_%.1f_.dat", distribs[d], stddev0, stddev1);
        
        cout << "output file: " << outputPathname << "\n";
        ofstream fout(outputPathname);
        if(!fout) {
            FATAL("error opening outputfile");
        }
        fout.precision(4);
        
        
        for(int n = 0; names[n]; n++) {
            //if(!active[n])
            //continue;
            
            vector<PairIdDistance> topIdDists;
            bool initializeIdDists = true;
            
            cout << "processing dataset:" << names[n] << " ...\n";
            sprintf(inputPathname, "udatasets/%s/%s_%s_mixed_%.1f_%.1f_.dat", names[n], names[n],distribs[d], stddev0, stddev1);
            
            
            cout << "input file: " << inputPathname << "\n";
            
            TimeSeriesCollection db(inputPathname, d+1);
            
            if(initializeIdDists) { // do it just the first time.
                db.computeTopKthresholds(10, topIdDists);
                initializeIdDists = false;
            }
            
            // allocating all methods ...
            Euclidean euclideanTrue(db, true);
            Euclidean euclidean(db);
            MovingEuclidean movingAverage(db,VARIANT_MOVINGAVG);
            MovingEuclidean exponentialMovingAverage(db, VARIANT_EXPMOVINGAVG);
            MovingEuclidean probabilisticMovingAverage(db, VARIANT_P_MOVINGAVG);
            MovingEuclidean probabilisticExponentialMovingAverage(db, VARIANT_P_EXPMOVINGAVG);
            
            // precision and recall for all methods ...
            PrecisionRecall prEuclidean("Euclidean");
            PrecisionRecall prDUST("DUST");
            PrecisionRecall prMovingAverage("MovingAverage");
            PrecisionRecall prExponentialMovingAverage("ExponentialMovingAverage");
            PrecisionRecall prProbabilisticMovingAverage("ProbabilisticMovingAverage");
            PrecisionRecall prProbabilisticExponentialMovingAverage("ProbabilisticExponentialMovingAverage");
            
            // now, compute precision,recall and avg query time.
            for(unsigned int i = 0; i < db.sequences.size(); i++) {
                TimeSeries &query = db.sequences[i];
                
                vector<int> exact = euclideanTrue.rangeQuery(query, topIdDists[query.id-1].second);
                vector<int> r;
                
                double euclideanThreshold =euclidean.distance(query, db.sequences[topIdDists[query.id-1].first-1]);
                double movingAverageThreshold =movingAverage.distance(query, db.sequences[topIdDists[query.id-1].first-1]);
                double exponentialMovingAverageThreshold =exponentialMovingAverage.distance(query, db.sequences[topIdDists[query.id-1].first-1]);
                double probabilisticMovingAverageThreshold = probabilisticMovingAverage.distance(query, db.sequences[topIdDists[query.id-1].first-1]);
                double probabilisticExponentialMovingAverageThreshold = probabilisticExponentialMovingAverage.distance(query, db.sequences[topIdDists[query.id-1].first-1]);
                
                
                // Euclidean
                prEuclidean.addStartTime();
                r = euclidean.rangeQuery(query, euclideanThreshold);
                prEuclidean.addStopTime();
                prEuclidean.add(exact, r);
                
                
                r.clear();


                // MovingAverage
                prMovingAverage.addStartTime();
                r = movingAverage.rangeQuery(query, movingAverageThreshold);
                prMovingAverage.addStopTime();
                prMovingAverage.add(exact, r);
                
                // ExponentialMovingAverage
                prExponentialMovingAverage.addStartTime();
                r = exponentialMovingAverage.rangeQuery(query, exponentialMovingAverageThreshold);
                prExponentialMovingAverage.addStopTime();
                prExponentialMovingAverage.add(exact, r);
                
                // ProbabilisticMovingAverage
                prProbabilisticMovingAverage.addStartTime();
                r = probabilisticMovingAverage.rangeQuery(query, probabilisticMovingAverageThreshold);
                prProbabilisticMovingAverage.addStopTime();
                prProbabilisticMovingAverage.add(exact, r);
                
                // ProbabilisticExponentialMovingAverage
                prProbabilisticExponentialMovingAverage.addStartTime();
                r = probabilisticExponentialMovingAverage.rangeQuery(query, probabilisticExponentialMovingAverageThreshold);
                prProbabilisticExponentialMovingAverage.addStopTime();
                prProbabilisticExponentialMovingAverage.add(exact, r);
                
                
            }
            
            // write to file the F1 scores...
            fout << names[n] << " " <<
                    prEuclidean.getF1() << " " <<
                    prDUST.getF1() << " " <<
                    prMovingAverage.getF1() << " " <<
                    prExponentialMovingAverage.getF1() << " " <<
                    prProbabilisticMovingAverage.getF1() << " " <<
                    prProbabilisticExponentialMovingAverage.getF1() << " " <<
            

            fout.flush();
            
        }
        
        
        fout.flush();
        fout.close();
        
    }
}

void experiment7_eval(double stddev0, double stddev1, double lambda, int w, ofstream *f);

void experiment7() {

    char outputPathname[1024];

    sprintf(outputPathname, "results/exp7.dat");
    cout << "output file: " << outputPathname << "\n";
    ofstream fout(outputPathname);
    if(!fout) {
        FATAL("error opening outputfile");
    }
    fout.precision(4);


    for(int w = 0; w <= 21; w+=2) {
        experiment7_eval(0.4, 1, .1, w, &fout);
        experiment7_eval(0.4, 1, 1, w, &fout);
    }

    for(double lambda = 0; lambda <= 1.01; lambda+= 0.1) {
        experiment7_eval(0.4, 1, lambda, 5, &fout);
        experiment7_eval(0.4, 1, lambda, 10, &fout);
    }

    fout.close();

}

void experiment7_eval(double stddev0, double stddev1, double lambda, int w, ofstream *fout) {
    
    int d = 1; // normal distrib.
    char inputPathname[1024];




    for(int n = 0; names[n]; n++) {
        //if(!active[n])
        //continue;

        vector<PairIdDistance> topIdDists;
        bool initializeIdDists = true;

        cout << "processing dataset:" << names[n] << " ...\n";
        sprintf(inputPathname, "udatasets/%s/%s_%s_mixed_%.1f_%.1f_.dat", names[n], names[n],distribs[d], stddev0, stddev1);


        cout << "input file: " << inputPathname << "\n";



        TimeSeriesCollection db(inputPathname, d+1);

        if(initializeIdDists) { // do it just the first time.
            db.computeTopKthresholds(10, topIdDists);
            initializeIdDists = false;
        }

        Euclidean euclideanTrue(db, true);

        MovingEuclidean probabilisticMovingAverage(db, VARIANT_P_MOVINGAVG);
        probabilisticMovingAverage.w = w;
        probabilisticMovingAverage.l = lambda;

        MovingEuclidean probabilisticExponentialMovingAverage(db, VARIANT_P_EXPMOVINGAVG);
        probabilisticExponentialMovingAverage.w = w;
        probabilisticExponentialMovingAverage.l = lambda;

        PrecisionRecall prProbabilisticMovingAverage("ProbabilisticMovingAverage");
        PrecisionRecall prProbabilisticExponentialMovingAverage("ProbabilisticExponentialMovingAverage");

        // now, compute precision,recall and avg query time.
        for(unsigned int i = 0; i  < db.sequences.size(); i++) {
            TimeSeries &query = db.sequences[i];

            vector<int> exact = euclideanTrue.rangeQuery(query, topIdDists[query.id-1].second);
            vector<int> r;

            double probabilisticMovingAverageThreshold = probabilisticMovingAverage.distance(query, db.sequences[topIdDists[query.id-1].first-1]);
            double probabilisticExponentialMovingAverageThreshold = probabilisticExponentialMovingAverage.distance(query, db.sequences[topIdDists[query.id-1].first-1]);


            // ProbabilisticMovingAverage
            prProbabilisticMovingAverage.addStartTime();
            r = probabilisticMovingAverage.rangeQuery(query, probabilisticMovingAverageThreshold);
            prProbabilisticMovingAverage.addStopTime();
            prProbabilisticMovingAverage.add(exact, r);

            // ProbabilisticExponentialMovingAverage
            prProbabilisticExponentialMovingAverage.addStartTime();
            r = probabilisticExponentialMovingAverage.rangeQuery(query, probabilisticExponentialMovingAverageThreshold);
            prProbabilisticExponentialMovingAverage.addStopTime();
            prProbabilisticExponentialMovingAverage.add(exact, r);

        }

        // write to file the F1 scores...
        *fout << names[n] << " " << lambda << " " << w << " " <<
                prProbabilisticMovingAverage.getF1() << " " <<
                prProbabilisticExponentialMovingAverage.getF1() << "\n";
        fout->flush();
    }




}


void  initOpt(int argc, char **argv) {
    int            c;
    
    o.synthetic = false;
    o.syntheticLength = -1;
    o.rfileCollection = NULL;
    o.rfileQuery = NULL;
    o.rfileQuery2 = NULL;
    o.wfile = NULL;
    
    if (argc ==1)
        help();
    
    opterr = 0;
    
    while ((c = getopt(argc, argv, "C:Q:q:w:")) != EOF)
        switch (c)
        {
        
        
        case 'C':
        SAFE_FREE(o.rfileCollection);
        o.rfileCollection = strdup(optarg);
        break;
        
        case 'q':
        SAFE_FREE(o.rfileQuery);
        o.rfileQuery = strdup(optarg);
        break;
        
        case 'Q':
        SAFE_FREE(o.rfileQuery2);
        o.rfileQuery2 = strdup(optarg);
        break;
        
        case 'w':
        SAFE_FREE(o.wfile);
        o.wfile = strdup(optarg);
        break;
        
        default:
        FATAL("option '" << (char) optopt << "' invalid");
    }
    
    
    
}

void help() {
    
}

void cleanUp() {
    SAFE_FREE(o.rfileCollection);
    SAFE_FREE(o.rfileQuery);
    SAFE_FREE(o.rfileQuery2);
    SAFE_FREE(o.wfile);
}
